import { of, Observable, Subject, from } from 'rxjs';
import { InjectionToken, Injectable, Inject, NgModule, Directive, Input, TemplateRef, ViewContainerRef, Optional, SkipSelf, Host, defineInjectable, inject } from '@angular/core';
import { Router } from '@angular/router';
import { switchMap, take, takeUntil, map, tap } from 'rxjs/operators';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const ACCESS_CONFIG = new InjectionToken('ACCESS_CONFIG');

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 */
class AccessStrategy {
}
class FakeAccessStrategy extends AccessStrategy {
    /**
     * @param {?} access
     * @return {?}
     */
    has(access) {
        return of(false);
    }
}
FakeAccessStrategy.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} config
 * @param {?=} __1
 * @return {?}
 */
function flatten(config, { parse = (/**
 * @param {?} v
 * @return {?}
 */
v => v), group = false } = {}) {
    /** @type {?} */
    const flatConfig = {};
    /**
     * @param {?} path
     * @param {?} value
     * @return {?}
     */
    function setConfig(path, value) {
        if (!flatConfig[path]) {
            flatConfig[path] = new Set();
        }
        flatConfig[path].add(value);
    }
    /**
     * @param {?} path
     * @param {?} delimiter
     * @param {?} prop
     * @return {?}
     */
    function getPath(path, delimiter, prop) {
        return path
            ? path + delimiter + prop
            : prop;
    }
    /**
     * @param {?} accesses
     * @param {?} prop
     * @param {?} value
     * @param {?} path
     * @return {?}
     */
    function visitor(accesses, prop, value, path) {
        if (Array.isArray(value)) {
            value.forEach((/**
             * @param {?} access
             * @return {?}
             */
            access => {
                visitor(accesses, prop, access, path);
            }));
        }
        else if (typeof value === 'object') {
            /** @type {?} */
            const childrenAccesses = children(value, getPath(path, '.', prop));
            if (group) {
                accesses = accesses.concat(childrenAccesses);
                accesses.forEach((/**
                 * @param {?} access
                 * @return {?}
                 */
                access => {
                    setConfig(getPath(path, ':', access.prop), access.action);
                }));
            }
        }
        else {
            /** @type {?} */
            const expression = parse(value.replace(/\s/g, ''));
            setConfig(getPath(path, ':', prop), expression);
            accesses = accesses.concat({ action: expression, prop });
        }
        return accesses;
    }
    /**
     * @param {?} obj
     * @param {?=} path
     * @return {?}
     */
    function children(obj, path = '') {
        return Object.keys(obj)
            .reduce((/**
         * @param {?} accesses
         * @param {?} prop
         * @return {?}
         */
        (accesses, prop) => visitor(accesses, prop, obj[prop], path)), []);
    }
    children(config);
    return flatConfig;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const TokenType = {
    PAR_OPEN: '('.charCodeAt(0),
    PAR_CLOSE: ')'.charCodeAt(0),
    OP_NOT: '!'.charCodeAt(0),
    BINARY_AND: '&'.charCodeAt(0),
    BINARY_OR: '|'.charCodeAt(0),
    LITERAL: 'LITERAL',
    END: 'END',
    LEAF: 'LEAF',
    ATOMIC: 'ATOMIC'
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const operator = (/**
 * @param {?} evaluator$
 * @param {?} op
 * @return {?}
 */
(evaluator$, op) => (/**
 * @param {?} source
 * @return {?}
 */
(source) => {
    return Observable.create((/**
     * @param {?} observer
     * @return {?}
     */
    (observer) => {
        /** @type {?} */
        const finish$ = new Subject();
        /** @type {?} */
        let lastEval;
        finish$
            .pipe(take(1))
            .subscribe((/**
         * @param {?} value
         * @return {?}
         */
        (value) => {
            observer.next(value);
            observer.complete();
        }), (/**
         * @param {?} e
         * @return {?}
         */
        e => observer.error(e)));
        source.
            pipe(switchMap(evaluator$), takeUntil(finish$))
            .subscribe((/**
         * @param {?} value
         * @return {?}
         */
        value => {
            lastEval = op === TokenType.OP_NOT ? !value : value;
            if (op === TokenType.BINARY_AND && !lastEval) {
                finish$.next(false);
            }
            else if (op === TokenType.BINARY_OR && lastEval) {
                finish$.next(true);
            }
        }), (/**
         * @param {?} e
         * @return {?}
         */
        e => {
            finish$.error(e);
        }), (/**
         * @param {?} _
         * @return {?}
         */
        _ => {
            finish$.next(lastEval);
        }));
    }));
}));

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const Tokenizer = (/**
 * @param {?} exp
 * @return {?}
 */
exp => {
    /** @type {?} */
    let literal = '';
    /** @type {?} */
    const tokens = [];
    for (const char of exp) {
        /** @type {?} */
        const code = char.charCodeAt(0);
        switch (code) {
            case TokenType.PAR_OPEN:
            case TokenType.PAR_CLOSE:
            case TokenType.OP_NOT:
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (literal) {
                    tokens.push({
                        type: TokenType.LITERAL,
                        value: literal
                    });
                    literal = '';
                }
                tokens.push({
                    type: code,
                    value: char
                });
                break;
            default:
                literal += char;
        }
    }
    if (literal)
        tokens.push({
            type: TokenType.LITERAL,
            value: literal
        });
    return tokens;
});

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const PolishNotation = (/**
 * @param {?} tokens
 * @return {?}
 */
tokens => {
    /** @type {?} */
    let queue = [];
    /** @type {?} */
    const stack = [];
    /** @type {?} */
    let subQueue = [];
    /** @type {?} */
    let subStack = [];
    tokens.forEach((/**
     * @param {?} token
     * @return {?}
     */
    token => {
        switch (token.type) {
            case TokenType.LITERAL:
            case TokenType.OP_NOT:
                if (subStack.length) {
                    subQueue.push(token);
                }
                else {
                    queue.push(token);
                }
                break;
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (subStack.length) {
                    subStack.unshift(token);
                }
                else {
                    stack.unshift(token);
                }
                break;
            case TokenType.PAR_OPEN:
                subStack.unshift(token);
                break;
            case TokenType.PAR_CLOSE:
                while (subStack.length &&
                    subStack[subStack.length - 1].type !== TokenType.PAR_OPEN) {
                    subQueue.unshift(subStack.pop());
                }
                subStack.pop();
                if (subStack.length && subStack[subStack.length - 1].type === TokenType.OP_NOT) {
                    subQueue.unshift(subStack.pop());
                }
                queue = queue.concat([...subStack, ...subQueue]);
                subStack = [];
                subQueue = [];
                break;
            default:
                break;
        }
        // debugger;
    }));
    /** @type {?} */
    const result = (stack.length && [...stack, ...queue]) || queue;
    return result;
});
/** @type {?} */
const PolishGenerator = (/**
 * @param {?} polish
 * @return {?}
 */
function* (polish) {
    for (let index = 0; index < polish.length - 1; index++) {
        yield polish[index];
    }
    return polish[polish.length - 1];
});

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class ExpNode {
    /**
     * @param {?} op
     * @param {?} left
     * @param {?=} right
     * @param {?=} literal
     */
    constructor(op, left, right, literal) {
        this.op = op;
        this.left = left;
        this.right = right;
        this.literal = literal;
    }
    /**
     * @return {?}
     */
    isLeaf() {
        return this.op === TokenType.LEAF;
    }
    /**
     * @return {?}
     */
    isAtomic() {
        return (this.isLeaf() || (this.op === TokenType.OP_NOT && this.left.isLeaf()));
    }
    /**
     * @return {?}
     */
    getLiteralValue() {
        return this.literal;
    }
    /**
     * @param {?} left
     * @param {?} right
     * @return {?}
     */
    static CreateAnd(left, right) {
        return new ExpNode(TokenType.BINARY_AND, left, right);
    }
    /**
     * @param {?} exp
     * @return {?}
     */
    static CreateNot(exp) {
        return new ExpNode(TokenType.OP_NOT, exp);
    }
    /**
     * @param {?} left
     * @param {?} right
     * @return {?}
     */
    static CreateOr(left, right) {
        return new ExpNode(TokenType.BINARY_OR, left, right);
    }
    /**
     * @param {?} lit
     * @return {?}
     */
    static CreateLiteral(lit) {
        return new ExpNode(TokenType.LEAF, null, null, lit);
    }
}
/** @type {?} */
const make = (/**
 * @param {?} gen
 * @return {?}
 */
gen => {
    /** @type {?} */
    const data = gen.next().value;
    switch (data.type) {
        case TokenType.LITERAL:
            return ExpNode.CreateLiteral(data.value);
        case TokenType.OP_NOT:
            return ExpNode.CreateNot(make(gen));
        case TokenType.BINARY_AND: {
            /** @type {?} */
            const left = make(gen);
            /** @type {?} */
            const right = make(gen);
            return ExpNode.CreateAnd(left, right);
        }
        case TokenType.BINARY_OR: {
            /** @type {?} */
            const left = make(gen);
            /** @type {?} */
            const right = make(gen);
            return ExpNode.CreateOr(left, right);
        }
    }
    return null;
});

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const parser = (/**
 * @param {?} exp
 * @return {?}
 */
(exp) => {
    /** @type {?} */
    const tokens = Tokenizer(exp.replace(/\s/g, ''));
    /** @type {?} */
    const polish = PolishNotation(tokens);
    /** @type {?} */
    const gen = PolishGenerator(polish);
    /** @type {?} */
    const tree = make(gen);
    return tree;
});

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
let configurationAccess = {};
/** @type {?} */
let hasAccessStrategy = (/**
 * @return {?}
 */
() => of(false));
/**
 * @param {?} config
 * @return {?}
 */
function setConfigurationAccess(config) {
    configurationAccess = flatten(config, { parse: parser, group: true });
}
/**
 * @param {?} accessTest
 * @return {?}
 */
function setHasAccessStrategy(accessTest) {
    hasAccessStrategy = accessTest;
}
/**
 * @param {?} expression
 * @return {?}
 */
function parse(expression) {
    /** @type {?} */
    const arr = expression.replace(/\s/g, '').split(':');
    return {
        path: (arr[0] || ''),
        action: (arr[1] || '')
    };
}
/**
 * @param {?} accessExpression
 * @return {?}
 */
function canAccessExpression(accessExpression) {
    return of(accessExpression)
        .pipe(map((/**
     * @param {?} ae
     * @return {?}
     */
    ae => ae.replace(/\s/g, ''))), switchMap((/**
     * @param {?} ae
     * @return {?}
     */
    ae => nodeEvaluator$1(parser(ae), hasAccessStrategy))));
}
/**
 * @param {?} accessPath
 * @return {?}
 */
function canAccessPaths(accessPath) {
    /** @type {?} */
    const access = Array.isArray(accessPath)
        ? accessPath
        : [accessPath];
    return from(access)
        .pipe(map((/**
     * @param {?} ap
     * @return {?}
     */
    ap => ap.replace(/\s/g, ''))), operator((/**
     * @param {?} ap
     * @return {?}
     */
    ap => canAccessPath(ap)), TokenType.BINARY_OR));
}
/**
 * @param {?} path
 * @return {?}
 */
function canAccessPath(path) {
    try {
        /** @type {?} */
        const access = configurationAccess[path];
        if (!!access) {
            return from(access)
                .pipe(operator((/**
             * @param {?} tree
             * @return {?}
             */
            tree => nodeEvaluator$1(tree, hasAccessStrategy)), TokenType.BINARY_OR));
        }
        console.error(`Undefined path ${path}`);
        return of(false);
    }
    catch (e) {
        console.error(e);
        return of(false);
    }
}
/**
 * @param {?} tree
 * @param {?} literalEvaluator
 * @return {?}
 */
function nodeEvaluator$1(tree, literalEvaluator) {
    if (tree.isLeaf()) {
        console.log('Evaluating with ' + tree.getLiteralValue());
        return literalEvaluator(tree.getLiteralValue());
    }
    if (tree.op === TokenType.OP_NOT) {
        return of(tree.left)
            .pipe(operator((/**
         * @param {?} tree
         * @return {?}
         */
        tree => nodeEvaluator$1(tree, literalEvaluator)), TokenType.OP_NOT));
    }
    if (tree.op === TokenType.BINARY_OR) {
        return from([tree.left, tree.right])
            .pipe(operator((/**
         * @param {?} tree
         * @return {?}
         */
        tree => nodeEvaluator$1(tree, literalEvaluator)), TokenType.BINARY_OR));
    }
    if (tree.op === TokenType.BINARY_AND) {
        return from([tree.left, tree.right])
            .pipe(operator((/**
         * @param {?} tree
         * @return {?}
         */
        tree => nodeEvaluator$1(tree, literalEvaluator)), TokenType.BINARY_AND));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AccessService {
    /**
     * @param {?} config
     * @param {?} accessStrategy
     */
    constructor(config, accessStrategy) {
        setHasAccessStrategy((/**
         * @param {?} accessName
         * @return {?}
         */
        accessName => accessStrategy.has(accessName)));
        setConfigurationAccess(config.accesses || {});
    }
    /**
     * @param {?} accessPaths
     * @return {?}
     */
    can(accessPaths) {
        return canAccessPaths(accessPaths);
    }
    /**
     * @param {?} accessExpression
     * @return {?}
     */
    canExpression(accessExpression) {
        return canAccessExpression(accessExpression);
    }
}
AccessService.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root'
            },] }
];
/** @nocollapse */
AccessService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] },
    { type: AccessStrategy }
];
/** @nocollapse */ AccessService.ngInjectableDef = defineInjectable({ factory: function AccessService_Factory() { return new AccessService(inject(ACCESS_CONFIG), inject(AccessStrategy)); }, token: AccessService, providedIn: "root" });

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AccessDirective {
    /**
     * @param {?} template
     * @param {?} viewContainer
     * @param {?} accessService
     * @param {?} parentAccessDirective
     */
    constructor(template, viewContainer, accessService, parentAccessDirective) {
        this.template = template;
        this.viewContainer = viewContainer;
        this.accessService = accessService;
        this.parentAccessDirective = parentAccessDirective;
        this.onDestroy$ = new Subject();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        if (this.template) {
            /** @type {?} */
            let ngxAccess = this.ngxAccess;
            if (this.parentAccessDirective) {
                const { path, action } = parse(this.parentAccessDirective.ngxAccess);
                if (this.ngxAccess) {
                    if (Array.isArray(this.ngxAccess)) {
                        ngxAccess = `${this.ngxAccess.map((/**
                         * @param {?} access
                         * @return {?}
                         */
                        access => access.replace('$', path)))}:${action}`;
                    }
                    else {
                        const { path: childPath, action: childAction } = parse(ngxAccess);
                        ngxAccess = `${childPath.replace('$', path)}:${childAction ? childAction : action}`;
                    }
                }
                else {
                    ngxAccess = `${path}:${action}`;
                }
            }
            this.accessService.can(ngxAccess)
                .pipe(takeUntil(this.onDestroy$))
                .subscribe((/**
             * @param {?} access
             * @return {?}
             */
            access => access
                ? this.viewContainer.createEmbeddedView(this.template)
                : this.ngxAccessElse
                    ? this.viewContainer.createEmbeddedView(this.ngxAccessElse)
                    : null));
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.onDestroy$.next();
        this.onDestroy$.complete();
    }
}
AccessDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxAccess]'
            },] }
];
/** @nocollapse */
AccessDirective.ctorParameters = () => [
    { type: TemplateRef, decorators: [{ type: Optional }] },
    { type: ViewContainerRef },
    { type: AccessService },
    { type: AccessDirective, decorators: [{ type: Optional }, { type: SkipSelf }, { type: Host }] }
];
AccessDirective.propDecorators = {
    ngxAccess: [{ type: Input }],
    ngxAccessElse: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AccessExpressionDirective {
    /**
     * @param {?} template
     * @param {?} viewContainer
     * @param {?} accessService
     */
    constructor(template, viewContainer, accessService) {
        this.template = template;
        this.viewContainer = viewContainer;
        this.accessService = accessService;
        this.onDestroy$ = new Subject();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        if (this.template) {
            this.accessService.canExpression(this.ngxAccessExpr)
                .pipe(takeUntil(this.onDestroy$))
                .subscribe((/**
             * @param {?} access
             * @return {?}
             */
            access => access
                ? this.viewContainer.createEmbeddedView(this.template)
                : this.ngxAccessExprElse
                    ? this.viewContainer.createEmbeddedView(this.ngxAccessExprElse)
                    : null));
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.onDestroy$.next();
        this.onDestroy$.complete();
    }
}
AccessExpressionDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxAccessExpr]'
            },] }
];
/** @nocollapse */
AccessExpressionDirective.ctorParameters = () => [
    { type: TemplateRef },
    { type: ViewContainerRef },
    { type: AccessService }
];
AccessExpressionDirective.propDecorators = {
    ngxAccessExpr: [{ type: Input }],
    ngxAccessExprElse: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AccessGuard {
    /**
     * @param {?} router
     * @param {?} accessService
     * @param {?} config
     */
    constructor(router, accessService, config) {
        this.router = router;
        this.accessService = accessService;
        this.config = config;
    }
    /**
     * @param {?} next
     * @return {?}
     */
    canActivate(next) {
        return this.accessService.can(next.data.accesses)
            .pipe(tap((/**
         * @param {?} hasAccess
         * @return {?}
         */
        hasAccess => !hasAccess && this.router.navigate([next.data.redirect || this.config.redirect]))));
    }
}
AccessGuard.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root'
            },] }
];
/** @nocollapse */
AccessGuard.ctorParameters = () => [
    { type: Router },
    { type: AccessService },
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] }
];
/** @nocollapse */ AccessGuard.ngInjectableDef = defineInjectable({ factory: function AccessGuard_Factory() { return new AccessGuard(inject(Router), inject(AccessService), inject(ACCESS_CONFIG)); }, token: AccessGuard, providedIn: "root" });

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class AccessModule {
    /**
     * @param {?} config
     * @return {?}
     */
    static forRoot(config) {
        return {
            ngModule: AccessModule,
            providers: [
                {
                    provide: ACCESS_CONFIG,
                    useValue: {
                        accesses: config.accesses || {},
                        redirect: config.redirect || '/unauthorized',
                    }
                },
                config.strategy || {
                    provide: AccessStrategy,
                    useClass: FakeAccessStrategy
                }
            ]
        };
    }
}
AccessModule.decorators = [
    { type: NgModule, args: [{
                declarations: [AccessDirective, AccessExpressionDirective],
                exports: [AccessDirective, AccessExpressionDirective]
            },] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { AccessModule, AccessDirective, AccessExpressionDirective, setConfigurationAccess, setHasAccessStrategy, parse, canAccessExpression, canAccessPaths, AccessGuard, AccessService, AccessStrategy, FakeAccessStrategy, ACCESS_CONFIG, AccessExpressionDirective as ɵe, AccessDirective as ɵa, AccessStrategy as ɵc, FakeAccessStrategy as ɵd, AccessService as ɵb };

//# sourceMappingURL=ngx-access.js.map